# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from __future__ import print_function
from .log import style, logger, logger_handler
from .time_utils import format_unixts
from .utils import compact_output, MAXFD
from .config import configuration
import errno, fcntl, logging, os, select, \
  signal, sys, threading, time, traceback, \
  subprocess, resource, heapq
if sys.version_info >= (3,):
    import queue, _thread
else:
    import Queue as queue, thread as _thread

# assuming import of this module is triggered from "the main" thread,
# the following call is intended to workaround python issue #7980
# http://bugs.python.org/issue7980
# (see also http://bugs.python.org/issue11108)
try:
    time.strptime("1970-01-01 00:00:00", "%Y-%m-%d %H:%M:%S")
except:
    pass

# max number of bytes read when reading asynchronously from a pipe
try:
    _MAXREAD = int(subprocess.Popen(["getconf", "_POSIX_SSIZE_MAX"], stdout=subprocess.PIPE, universal_newlines=True).communicate()[0])
except:
    _MAXREAD = 32767

if sys.platform.startswith('darwin') or sys.platform.startswith('win'):

    # minimal abstraction of poll over select allowing the conductor
    # to run on platforms lacking poll: macosx, windows. poll is
    # removed from python on macos due to an unclear (buggy or
    # unusual?) behavior.  this *minimal* abstraction only goes as far
    # as needed to allow the conductor to run, this is probably *not*
    # a full emulation of poll over select.

    POLLIN=1
    POLLPRI=2
    POLLOUT=4
    POLLERR=8
    POLLHUP=16
    POLLNVAL=32

    def dict_item_and(dic, item, field):
        if item in dic:
            dic[item] |= field
        else:
            dic[item] = field

    class poll(object):

        def __init__(self):
            self._readfds = list()
            self._writefds = list()
            self._exceptfds = list()

        def register(self, fd, eventmask = POLLIN | POLLPRI | POLLOUT):
            if eventmask & POLLIN or eventmask & POLLPRI:
                readfds_set = set(self._readfds)
                readfds_set.add(fd)
                self._readfds = list(readfds_set)
            if eventmask & POLLOUT:
                writefds_set = set(self._writefds)
                writefds_set.add(fd)
                self._writefds = list(writefds_set)
            if eventmask & POLLIN or eventmask & POLLOUT:
                exceptfds_set = set(self._exceptfds)
                exceptfds_set.add(fd)
                self._exceptfds = list(exceptfds_set)

        def unregister(self, fd):
            try:
                self._readfds.remove(fd)
            except ValueError:
                pass
            try:
                self._writefds.remove(fd)
            except ValueError:
                pass
            try:
                self._exceptfds.remove(fd)
            except ValueError:
                pass

        def poll(self, timeout = None):
            if timeout != None: timeout = timeout / 1000.0
            (ready_readfds,
             ready_writefds,
             ready_exceptfds) = select.select(
                self._readfds,
                self._writefds,
                self._exceptfds, timeout)
            fd_events = dict()
            for fd in ready_readfds: dict_item_and(fd_events, fd, POLLIN)
            for fd in ready_writefds: dict_item_and(fd_events, fd, POLLOUT)
            for fd in ready_exceptfds: dict_item_and(fd_events, fd, POLLERR)
            return list(fd_events.items())

else:
    from select import poll, POLLIN, POLLPRI, POLLOUT, POLLERR, POLLHUP, POLLNVAL

def _event_desc(event):
    """For debugging: user friendly representation of the event bitmask returned by poll()."""
    desc = ""
    first = True
    for t in ('POLLIN', 'POLLPRI', 'POLLOUT', 'POLLERR', 'POLLHUP', 'POLLNVAL'):
        if event & globals()[t]:
            if not first: desc += '|'
            desc += t
            first = False
    return desc

def _checked_waitpid(pid, options):
    """Wrapper for waitpid.

    - restarts if interrupted by signal

    - returns (0, 0) if no child processes

    - returns the waitpid tuple containing process id and exit status
      indication otherwise.
    """
    while True:
        try:
            return os.waitpid(pid, options)
        except OSError as e:
            if e.errno == errno.ECHILD:
                return 0, 0
            elif e.errno == errno.EINTR:
                continue
            else:
                raise

def _read_asmuch(fileno):
    """Read as much as possible from a file descriptor without blocking.

    Relies on the file descriptor to have been set non blocking.

    Returns a tuple (string, eof). string is the data read, eof is
    a boolean flag.
    """
    buf = b''
    eof = False
    while True:
        try:
            tmpbuf = os.read(fileno, _MAXREAD)
        except OSError as err:
            if err.errno == errno.EAGAIN:
                break
            else:
                raise
        if len(tmpbuf) == 0:
            eof = True
            break
        else:
            buf += tmpbuf
    return (buf, eof)

def _set_fd_nonblocking(fileno):
    """Sets a file descriptor in non blocking mode.

    Returns the previous state flags.
    """
    status_flags = fcntl.fcntl(fileno, fcntl.F_GETFL, 0)
    fcntl.fcntl(fileno, fcntl.F_SETFL, status_flags | os.O_NONBLOCK)
    return status_flags

def remove_from_heapq(hq, pred):
    # remove all objects satisfying predicate pred from heapq hq in O(log(n))
    i = 0
    while i<len(hq):
        if pred(hq[i]):
            hq[i] = hq[-1]
            hq.pop()
            if i < len(hq):
                heapq._siftup(hq, i)
            continue
        i += 1

class _Conductor(object):

    """Manager of the subprocesses outputs and lifecycle.

    There must be **one and only one** instance of this class

    Instance of _Conductor will start two threads, one for handling
    asynchronously subprocesses outputs and part of the process
    lifecycle management (the main 'conductor' thread), and a second
    thread (the 'reaper' thread) for handling asynchronously
    subprocess terminations.
    """

    def __init__(self):
        self.lock = threading.RLock()
        self.condition = threading.Condition(self.lock)
        # this lock and conditions are used for:
        #
        # - mutual exclusion and synchronization beetween sections of
        # code in _Conductor.__io_loop (conductor thread), and in
        # Process.start() and Process.wait() (main thread)
        #
        # - mutual exclusion beetween sections of code in
        #   _Conductor.__io_loop() (conductor thread) and in
        #   _Conductor.__reaper_thread_func() (reaper thread)

        self.__io_thread = threading.Thread(target = self.__io_thread_func, name = "I/O")
        self.__io_thread.setDaemon(True)
        # thread will terminate automatically when the main thread
        # exits.  once in a while, this can trigger an exception, but
        # this seems to be safe and to be related to this issue:
        # http://bugs.python.org/issue1856
        self.__rpipe, self.__wpipe = os.pipe()  # pipe used to wakeup
                                                # the conductor thread
                                                # from the main thread
                                                # when needed
        _set_fd_nonblocking(self.__rpipe)   # the reading function
                                            # _read_asmuch() relies on
                                            # file descriptors to be non
                                            # blocking
        _set_fd_nonblocking(self.__wpipe)   # because we call
                                            # signal.set_wakeup_fd on this pipe
        self.__poller = poll()   # asynchronous I/O with all
                                 # subprocesses filehandles
        self.__poller.register(self.__rpipe,
                               POLLIN
                               | POLLERR)
        self.__processes = set()    # the set of `Process` handled by
                                    # this `_Conductor`
        self.__fds = dict() # keys: the file descriptors currently polled by
                            # this `_Conductor`
                            #
                            # values: tuples (`Process`, `Process`'s
                            # function to handle activity for this
                            # descriptor)
        self.__pids = dict()    # keys: the pids of the subprocesses
                                # launched by this `_Conductor`
                                #
                                # values: their `Process`
        self.__timeline = [] # heapq of `Process` with a timeout date
        self.__process_actions = queue.Queue()
                                # thread-safe FIFO used to send requests
                                # from main thread and conductor thread:
                                # we enqueue tuples (function to call,
                                # tuple of parameters to pass to this
                                # function))
        self.__reaper_thread_running = False
                                # to keep track wether reaper thread is
                                # running
        signal.set_wakeup_fd(self.__wpipe)
        self.pgrp = self.__start_pgrp()

    def __str__(self):
        return "<" + style.object_repr("Conductor") + "(num processes=%i, num fds=%i, num pids=%i, timeline length=%i)>" % (len(self.__processes), len(self.__fds), len(self.__pids), len(self.__timeline))

    def __start_pgrp(self):
        # start a dedicated dummy process, having its own process
        # group, in order to group all processes handled by this
        # conductor
        ppid = os.getpid()
        pid = os.fork()
        if pid == 0:
            os.setpgid(0, 0)
            os.chdir("/")
            os.umask(0)
            maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
            if (maxfd == resource.RLIM_INFINITY):
                maxfd = MAXFD
            for fd in range(0, maxfd):
                try: os.close(fd)
                except OSError: pass
            while True:
                # poll each second that my parent is still alive. If
                # not, die. Optionnaly kill all instanciated childs.
                if os.getppid() != ppid:
                    if configuration['kill_childs_at_end']:
                        os.killpg(0, signal.SIGTERM)
                    os._exit(0)
                time.sleep(1)
        else:
            return pid

    def __wakeup(self):
        # wakeup the I/O thread
        os.write(self.__wpipe, b'.')

    def start(self):
        """Start the conductor thread."""
        self.__io_thread.start()
        return self

    def terminate(self):
        """Close the conductor thread."""
        # the closing of the pipe will wake the conductor which will
        # detect this closing and self stop
        logger.debug("terminating I/O thread of %s", self)
        os.close(self.__wpipe)
        self.__io_thread.join()
        logger.debug("I/O thread of %s terminated", self)

    def start_process(self, process):
        """Register a new `execo.process.Process` to be started and handled by the conductor.

        Intended to be called from main thread.
        """
        self.__process_actions.put_nowait((self.__handle_start_process, (process,)))
        self.__wakeup()

    def update_process(self, process):
        """Update `execo.process.Process` to the conductor.

        Intended to be called from main thread.

        Currently: only update the force kill timeout.
        """
        self.__process_actions.put_nowait((self.__handle_update_process, (process,)))
        self.__wakeup()

    def remove_process(self, process, exit_code = None):
        """Remove a `execo.process.Process` from the conductor.

        Intended to be called from main thread.
        """
        self.__process_actions.put_nowait((self.__handle_remove_process, (process, exit_code)))
        self.__wakeup()

    def notify_process_terminated(self, pid, exit_code):
        """Tell the conductor thread that a `execo.process.Process` has terminated.

        Intended to be called from the reaper thread.
        """
        self.__process_actions.put_nowait((self.__handle_notify_process_terminated, (pid, exit_code)))
        self.__wakeup()

    def __handle_start_process(self, process):
        assert(not process.started and not process.ended)
        assert(process not in self.__processes)
        process._actual_start()
        if not process.ended:
            fileno_stdout = process.stdout_fd
            fileno_stderr = process.stderr_fd
            self.__processes.add(process)
            _set_fd_nonblocking(fileno_stdout)
            _set_fd_nonblocking(fileno_stderr)
            self.__fds[fileno_stdout] = (process, process._handle_stdout)
            self.__fds[fileno_stderr] = (process, process._handle_stderr)
            self.__poller.register(fileno_stdout,
                                   POLLIN
                                   | POLLERR)
            self.__poller.register(fileno_stderr,
                                   POLLIN
                                   | POLLERR)
            self.__pids[process.pid] = process
            if process.timeout_date != None:
                heapq.heappush(self.__timeline,(process.timeout_date, process))
            if self.__reaper_thread_running == False:
                self.__reaper_thread_running = True
                reaper_thread = threading.Thread(target = self.__reaper_thread_func, name = "Reaper")
                reaper_thread.setDaemon(True)
                reaper_thread.start()

    def __handle_update_process(self, process):
        # intended to be called from conductor thread
        # Currently: only update the force kill timeout.
        logger.fdebug("update force kill timeout of %s in %s", str(process), self)
        if process not in self.__processes:
            return  # this will frequently occur if the process kills
                    # quickly because the process will already be
                    # killed and reaped before __handle_update_process
                    # is called
        if process._force_kill_timeout_date != None:
            heapq.heappush(self.__timeline,(process._force_kill_timeout_date, process))

    def __handle_remove_process(self, process, exit_code = None):
        # intended to be called from conductor thread
        # unregister a Process from conductor
        logger.fdebug("removing %s from %s", str(process), self)
        if process not in self.__processes:
            raise ValueError("trying to remove a process which was not yet added to conductor")
        remove_from_heapq(self.__timeline, lambda e: e[1] == process)
        del self.__pids[process.pid]
        fileno_stdout = process.stdout_fd
        fileno_stderr = process.stderr_fd
        last_bytes = b''
        if fileno_stdout in self.__fds:
            del self.__fds[fileno_stdout]
            self.__poller.unregister(fileno_stdout)
            # read the last data that may be available on stdout of
            # this process
            try:
                (last_bytes, _) = _read_asmuch(fileno_stdout)
            except OSError as e:
                if e.errno == errno.EBADF: last_bytes = b''
                else: raise e
        process._handle_stdout(last_bytes, True, False)
        last_bytes = b''
        if fileno_stderr in self.__fds:
            del self.__fds[fileno_stderr]
            self.__poller.unregister(fileno_stderr)
            # read the last data that may be available on stderr of
            # this process
            try:
                (last_bytes, _) = _read_asmuch(fileno_stderr)
            except OSError as e:
                if e.errno == errno.EBADF: last_bytes = b''
                else: raise e
        process._handle_stderr(last_bytes, True, False)
        self.__processes.remove(process)
        if exit_code != None:
            process._set_terminated(exit_code = exit_code)

    def __get_next_timeout(self):
        """Return the remaining time until the smallest timeout date of all registered `execo.process.Process`."""
        next_timeout = None
        if len(self.__timeline) > 0:
            next_timeout = self.__timeline[0][0] - time.time()
        return next_timeout

    def __check_timeouts(self):
        """Iterate all registered `execo.process.Process` whose timeout is reached, kill them.

        And remove them from the timeline.
        """
        now = time.time()
        while len(self.__timeline) > 0 and self.__timeline[0][0] <= now:
            _, process = heapq.heappop(self.__timeline)
            if process._force_kill_timeout_date != None and now >= process._force_kill_timeout_date:
                logger.debug("force kill timeout on %s" % (str(process),))
                process._force_kill()
            elif now >= process.timeout_date:
                logger.debug("timeout on %s" % (str(process),))
                process._timeout_kill()

    def __update_terminated_processes(self):
        """Ask operating system for all processes that have terminated, self remove them.

        Intended to be called from conductor thread.
        """
        exit_pid, exit_code = _checked_waitpid(- self.pgrp, os.WNOHANG)
        while exit_pid != 0:
            process = self.__pids[exit_pid]
            logger.fdebug("process pid %s terminated: %s", exit_pid, str(process))
            self.__handle_remove_process(process, exit_code)
            exit_pid, exit_code = _checked_waitpid(- self.pgrp, os.WNOHANG)

    def __handle_notify_process_terminated(self, pid, exit_code):
        # intended to be called from conductor thread
        # remove a process based on its pid (as reported by waitpid)
        process = self.__pids.get(pid)
        if process:
            self.__handle_remove_process(process, exit_code)

    def __remove_handle(self, fd):
        # remove a file descriptor both from our member(s) and from
        # the Poll object
        del self.__fds[fd]
        self.__poller.unregister(fd)

    def __io_thread_func(self):
        # wrapper around the actual io loop func for
        # exception handling
        try:
            self.__io_loop()
        except Exception: #IGNORE:W0703
            print("exception in conductor I/O loop thread")
            traceback.print_exc()
            _thread.interrupt_main()

    def __io_loop(self):
        # conductor thread infinite I/O loop
        finished = False
        while not finished:
            descriptors_events = []
            delay = self.__get_next_timeout()   # poll timeout will be
                                                # the delay until the
                                                # first of our
                                                # registered processes
                                                # reaches its timeout
            logger.fdebug("polling %i descriptors (+ rpipe) with timeout %s", len(self.__fds), "%.3fs" % delay if delay != None else "None")
            if delay == None or delay > 0: # don't even poll if poll timeout is <= 0
                if delay != None: delay *= 1000 # poll needs delay in millisecond
                descriptors_events = self.__poller.poll(delay)
            logger.fdebug("len(descriptors_events) = %i", len(descriptors_events))
            event_on_rpipe = None   # we want to handle any event on
                                    # rpipe after all other file
                                    # descriptors, hence this flag
            for descriptor_event in descriptors_events:
                fd = descriptor_event[0]
                event = descriptor_event[1]
                if fd == self.__rpipe:
                    event_on_rpipe = event
                else:
                    if fd in self.__fds:
                        process, stream_handler_func = self.__fds[fd]
                        logger.fdebug("event %s on fd %s, process %s", _event_desc(event), fd, str(process))
                        if event & POLLIN:
                            (string, eof) = _read_asmuch(fd)
                            stream_handler_func(string, False, False)
                            if eof:
                                self.__remove_handle(fd)
                        #if event & POLLHUP:
                        #    stream_handler_func('', True, False)
                        #    self.__remove_handle(fd)
                        if event & POLLERR:
                            stream_handler_func('', False, True)
                            self.__remove_handle(fd)
            self.__check_timeouts()
            if event_on_rpipe != None:
                logger.fdebug("event %s on inter-thread pipe", _event_desc(event_on_rpipe))
                if event_on_rpipe & POLLIN:
                    (string, eof) = _read_asmuch(self.__rpipe)
                    if eof:
                        # pipe closed -> auto stop the thread
                        finished = True
                if event_on_rpipe & POLLHUP:
                    finished = True
                if event_on_rpipe & POLLERR:
                    finished = True
                    raise IOError("Error on inter-thread communication pipe")
            with self.lock:
                while True:
                    try:
                        # call (in the right order!) all functions
                        # enqueued from other threads
                        func, args = self.__process_actions.get_nowait()
                    except queue.Empty:
                        break
                    func(*args)
                self.__update_terminated_processes()
                self.condition.notifyAll()
        logger.debug("conductor exiting I/O loop")
        self.__poller.unregister(self.__rpipe)
        try:
            os.close(self.__rpipe)
        except:
            pass
        try:
            os.close(self.__wpipe)
        except:
            pass

    def __reaper_thread_func(self):
        # run func for the reaper thread, whose role is to wait to be
        # notified by the operating system of terminated processes
        while True:
            exit_pid, exit_code = _checked_waitpid(- self.pgrp, 0)
            with self.lock:
                # this lock is needed to ensure that:
                #
                # - Conductor.__update_terminated_processes() won't be
                #   called before the process has been registered to
                #   the conductor
                #
                # - the following code cannot run while
                #    __handle_start_process() is running
                if (exit_pid, exit_code) == (0, 0):
                    if len(self.__processes) == 0:
                        # no more child processes, we stop this thread
                        # (another instance will be restarted as soon as
                        # another process is started)
                        self.__reaper_thread_running = False
                        break
                    # (exit_pid, exit_code) can be == (0, 0) and
                    # len(self.__processes) > 0 when and only when
                    # _checked_waitpid has returned because there are
                    # no more child and at the same time (but before
                    # entering the locked section)
                    # __handle_start_process() just added a process.
                else:
                    logger.debug("process with pid=%s terminated, exit_code=%s", exit_pid, exit_code)
                    self.notify_process_terminated(exit_pid, exit_code)

the_conductor = _Conductor().start()
"""The **one and only** `execo.conductor._Conductor` instance."""

#------------------------------------------------------------------------
#
# debug tools
#
#------------------------------------------------------------------------

def debug_dump_processes():
    with the_conductor.lock:
        print("\n===== %s dump current %i conductor handled processes:\n" % (format_unixts(time.time()), len(the_conductor._Conductor__processes),), file=sys.stderr)
        for process in the_conductor._Conductor__processes:
            print("=====", file=sys.stderr)
            print(str(process), file=sys.stderr)
            print("stdout:\n" + compact_output(process.stdout), file=sys.stderr)
            print("stderr:\n" + compact_output(process.stderr), file=sys.stderr)
            print(file=sys.stderr)
        print("=====\n", file=sys.stderr)

_debug_thread = None
_debug_thread_id = None
_debug_thread_lock = threading.Lock()

def debug_dump_threads():
    print("\n===== %s dump thread stack frames. %i threads. conductor lock = %s:\n" % (
        format_unixts(time.time()),
        len(sys._current_frames()),
        the_conductor.lock), file=sys.stderr)
    idx=0
    for thread_id, frame in sys._current_frames().items():
        print("===== thread #%i [%#x] refcount = %s" % (idx, thread_id, sys.getrefcount(frame)), file=sys.stderr)
        if thread_id != _debug_thread_id:
            traceback.print_stack(frame, file = sys.stderr)
        else:
            print("  debug thread, skipping", file=sys.stderr)
        idx += 1
    print("=====\n", file=sys.stderr)

class _DebugThread(threading.Thread):

    def __init__(self, interval = 30):
        super(_DebugThread, self).__init__(name = "debug")
        self.interval = interval
        self.terminate = threading.Event()
        self.setDaemon(True)

    def run(self):
        while True:
            if self.terminate.is_set():
                break
            debug_dump_threads()
            time.sleep(self.interval)

    def stop(self):
        self.terminate.set()

def debug_toggle_thread(interval = 30):

    global _debug_thread, _debug_thread_id

    with _debug_thread_lock:
        if _debug_thread:
            _debug_thread.stop()
            _debug_thread = None
            _debug_thread_id = None
        else:
            _debug_thread = _DebugThread(interval)
            _debug_thread.start()
            _debug_thread_id = _debug_thread.ident



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from .utils import comma_join
from .log import style

class Host(object):

    """A host to connect to.

    - Has an address (mandatory)

    - Can optionaly have a user, a keyfile, a port, which are used for
      remote connection and authentification (with a ssh like remote
      connection tool).

    Has an intuitive comparison and hashing behavior: two `execo.host.Host` with
    the same members (address, user, keyfile, port) will hash equally
    and will be seen as identical keys in a set or dict.

    >>> h1 = Host('localhost')
    >>> h1.user = 'root'
    >>> h1
    Host('localhost', user='root')
    >>> h2 = Host('localhost', user = 'root')
    >>> h1 == h2
    True
    >>> d = dict()
    >>> d[h1] = True
    >>> d[h2]
    True
    """

    def __init__(self, address, user = False, keyfile = False, port = False):
        """
        :param address: (string or `execo.host.Host`) the host address
          or another `execo.host.Host` instance which will be copied
          into this new instance

        :param user: (string) optional user whith which to connect. If
          False (default value), means use the default user. If None,
          means don't use any user.

        :param keyfile: (string) optional keyfile whith which to
          connect. If False (default value), means use the default
          keyfile. If None, means don't use any keyfile.

        :param port: (integer) optional port to which to connect. If
          False (default value), means use the default port. If None,
          means don't use any port.
        """
        if isinstance(address, Host):
            self.address = address.address
            """the host address"""
            self.user = address.user
            self.keyfile = address.keyfile
            self.port = address.port
        else:
            self.address = address
            self.user = None
            self.keyfile = None
            self.port = None
        if user != False: self.user = user
        """login name for remote connections"""
        if keyfile != False: self.keyfile = keyfile
        """keyfile to use for remote connections"""
        if port != False: self.port = port
        """tcp port to use for remote connection"""

    def __eq__(self, other):
        if not other or not isinstance(other, Host):
            return False
        return (self.address == other.address and
                self.user == other.user and
                self.keyfile == other.keyfile and
                self.port == other.port)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return (self.address.__hash__()
                ^ self.user.__hash__()
                ^ self.keyfile.__hash__()
                ^ self.port.__hash__())

    def _args(self):
        args = style.host(repr(self.address))
        if self.user: args = comma_join(args, "user=%r" % (self.user,))
        if self.keyfile: args = comma_join(args, "keyfile=%r" % (self.keyfile,))
        if self.port: args = comma_join(args, "port=%r" % (self.port,))
        return args

    def __repr__(self):
        return "Host(%s)" % (self._args())

def get_hosts_list(hosts):
    """Deep copy an iterable of `execo.host.Host` or hostnames to a list of `execo.host.Host`.

    order is preserved
    """
    return [ Host(host) for host in hosts ]

def get_unique_hosts_list(hosts):
    """Deep copy an iterable of `execo.host.Host` or hostnames to a list of unique `execo.host.Host`.

    order is preserved
    """
    tmphostsset = set()
    newhosts = []
    for h in get_hosts_list(hosts):
        if h not in tmphostsset:
            newhosts.append(h)
            tmphostsset.add(h)
    return newhosts

def get_hosts_set(hosts):
    """Deep copy an iterable of `execo.host.Host` or hostnames to a set of `execo.host.Host`."""
    return set(get_hosts_list(hosts))



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from .config import g5k_configuration
from execo.config import make_connection_params
from execo.exception import ProcessesFailed
from execo.host import Host
from execo.process import get_process
from execo.time_utils import get_unixts, sleep
from execo_g5k.config import default_frontend_connection_params
from execo_g5k.utils import get_frontend_host
from .oar import format_oar_date, format_oar_duration, _date_in_range, \
    oar_date_to_unixts, oar_duration_to_seconds
from .api_utils import get_g5k_sites, get_cluster_site
import os
import re

def _quote_hack(rdef):
    return rdef.replace('{', '{\\\\\\\\\\\\\\"').replace('}', '\\\\\\\\\\\\\\"}')

def get_oargridsub_commandline(job_specs, reservation_date = None,
                               walltime = None, job_type = None,
                               queue = None, directory = None,
                               additional_options = None):
    oargridsub_cmdline = 'oargridsub -v'
    key = g5k_configuration.get('oar_job_key_file')
    if key == None:
        key = os.environ.get('OAR_JOB_KEY_FILE')
    if key != None:
        oargridsub_cmdline += ' -i %s' % (key,)
    if reservation_date:
        oargridsub_cmdline += ' -s "%s"' % (format_oar_date(reservation_date),)
    if queue != None:
        oargridsub_cmdline += ' -q "%s"' % (queue,)
    if job_type != None:
        oargridsub_cmdline += ' -t "%s"' % (job_type,)
    if walltime != None:
        oargridsub_cmdline += ' -w "%s"' % (format_oar_duration(walltime),)
    if directory != None:
        oargridsub_cmdline += ' -d "%s"' % (directory,)
    if additional_options != None:
        oargridsub_cmdline += ' %s' % (additional_options,)
    firstclusteralias = True
    for (spec, clusteralias) in job_specs:
        if firstclusteralias:
            firstclusteralias = False
            oargridsub_cmdline += ' '
        else:
            oargridsub_cmdline += ','
        oargridsub_cmdline += '%s:rdef="%s"' % (clusteralias, _quote_hack(spec.resources))
        if spec.job_type != None:
            oargridsub_cmdline += ':type="%s"' % (spec.job_type,)
        if spec.sql_properties != None:
            oargridsub_cmdline += ':prop="%s"' % (spec.sql_properties,)
        if spec.name != None:
            oargridsub_cmdline += ':name="%s"' % (spec.name,)
    return oargridsub_cmdline

def oargridsub(job_specs, reservation_date = None,
               walltime = None, job_type = None,
               queue = None, directory = None,
               additional_options = None,
               frontend_connection_params = None,
               timeout = False):
    """Submit oargrid jobs.

    :param job_specs: iterable of tuples (OarSubmission,
      clusteralias). Reservation date, walltime, queue, directory,
      project, additional_options, command of the OarSubmission are
      ignored.

    :param reservation_date: grid job reservation date. Default: now.

    :param walltime: grid job walltime.

    :param job_type: type of job for all clusters: deploy, besteffort,
      cosystem, checkpoint, timesharing.

    :param queue: oar queue to use.

    :param directory: directory where the reservation will be
      launched.

    :param additional_options: passed directly to oargridsub on the
      command line.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use
      ``execo_g5k.config.g5k_configuration['default_timeout']``. None
      means no timeout.

    Returns a tuple (oargrid_job_id, ssh_key), or (None, None) if
    error.

    Note that, as oargrid does not handle correctly quoting sql
    clauses enclosed inside braces, this function tries to
    automatically overcome this limitation by adding some, with the
    right escaping (backslashes). Also, note that oargrid's command
    line parser does not handle correctly commas in sql clauses
    enclosed inside braces, as it considers it as a rdef
    separator. This prevents, for example, using comma separated list
    values for ``NOT IN`` clauses.
    """
    if isinstance(timeout, bool) and timeout == False:
        timeout = g5k_configuration.get('default_timeout')
    oargridsub_cmdline = get_oargridsub_commandline(job_specs, reservation_date,
                                                    walltime, job_type, queue,
                                                    directory, additional_options)
    process = get_process(oargridsub_cmdline,
                          host = get_frontend_host(),
                          connection_params = make_connection_params(frontend_connection_params,
                                                                     default_frontend_connection_params))
    process.timeout = timeout
    process.pty = True
    process.run()
    job_id = None
    ssh_key = None
    if process.ok:
        mo = re.search("^\[OAR_GRIDSUB\] Grid reservation id = (\d+)\s*$", process.stdout, re.MULTILINE)
        if mo != None:
            job_id = int(mo.group(1))
        mo = re.search("^\[OAR_GRIDSUB\] SSH KEY : (\S*)\s*$", process.stdout, re.MULTILINE)
        if mo != None:
            ssh_key = mo.group(1)
    if job_id != None:
        return (job_id, ssh_key)
    else:
        return (None, None)

def oargriddel(job_ids, frontend_connection_params = None, timeout = False):
    """Delete oargrid jobs.

    Ignores any error, so you can delete inexistant jobs, already
    deleted jobs, or jobs that you don't own. Those deletions will be
    ignored.

    :param job_ids: iterable of oar grid job ids.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use ``g5k_configuration['default_timeout']``. None means no
      timeout.
    """
    if isinstance(timeout, bool) and timeout == False:
        timeout = g5k_configuration.get('default_timeout')
    processes = []
    for job_id in job_ids:
        p = get_process("oargriddel %i" % (job_id,),
                        host = get_frontend_host(),
                        connection_params = make_connection_params(frontend_connection_params,
                                                                   default_frontend_connection_params))
        p.timeout = timeout
        p.nolog_exit_code = True
        p.pty = True
        processes.append(p)
    for process in processes: process.start()
    for process in processes: process.wait()

def get_current_oargrid_jobs(start_between = None,
                             end_between = None,
                             frontend_connection_params = None,
                             timeout = False):
    """Return a list of current active oargrid job ids.

    :param start_between: a tuple (low, high) of endpoints. Filters
      and returns only jobs whose start date is in between these
      endpoints.

    :param end_between: a tuple (low, high) of endpoints. Filters and
      returns only jobs whose end date is in between these endpoints.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use
      ``execo_g5k.config.g5k_configuration['default_timeout']``. None
      means no timeout.
    """
    if isinstance(timeout, bool) and timeout == False:
        timeout = g5k_configuration.get('default_timeout')
    if start_between: start_between = [ get_unixts(t) for t in start_between ]
    if end_between: end_between = [ get_unixts(t) for t in end_between ]
    process = get_process("oargridstat",
                          host = get_frontend_host(),
                          connection_params = make_connection_params(frontend_connection_params,
                                                                     default_frontend_connection_params))
    process.timeout = timeout
    process.pty = True
    process.run()
    if process.ok:
        jobs = re.findall("Reservation # (\d+):", process.stdout, re.MULTILINE)
        oargrid_job_ids = [ int(j) for j in jobs ]
        if start_between or end_between:
            filtered_job_ids = []
            for job in oargrid_job_ids:
                info = get_oargrid_job_info(job, timeout)
                if (_date_in_range(info['start_date'], start_between)
                    and _date_in_range(info['start_date'] + info['walltime'], end_between)):
                    filtered_job_ids.append(job)
            oargrid_job_ids = filtered_job_ids
        return oargrid_job_ids
    else:
        raise ProcessesFailed([process])

def get_oargrid_job_info(oargrid_job_id = None, frontend_connection_params = None, timeout = False):
    """Return a dict with informations about an oargrid job.

    :param oargrid_job_id: the oargrid job id.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use
      ``execo_g5k.config.g5k_configuration['default_timeout']``. None
      means no timeout.

    Hash returned contains these keys:

    - ``start_date``: unix timestamp of job's start date

    - ``walltime``: job's walltime in seconds

    - ``user``: job's user
    """
    if isinstance(timeout, bool) and timeout == False:
        timeout = g5k_configuration.get('default_timeout')
    process = get_process("oargridstat %i" % (oargrid_job_id,),
                          host = get_frontend_host(),
                          connection_params = make_connection_params(frontend_connection_params,
                                                                     default_frontend_connection_params))
    process.timeout = timeout
    process.pty = True
    process.run()
    job_info = dict()
    start_date_result = re.search("start date : (\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d)", process.stdout, re.MULTILINE)
    if start_date_result:
        start_date = oar_date_to_unixts(start_date_result.group(1))
        job_info['start_date'] = start_date
    walltime_result = re.search("walltime : (\d+:\d?\d:\d?\d)", process.stdout, re.MULTILINE)
    if walltime_result:
        walltime = oar_duration_to_seconds(walltime_result.group(1))
        job_info['walltime'] = walltime
    user_result = re.search("user : (\S+)", process.stdout, re.MULTILINE)
    if user_result:
        user = user_result.group(1)
        job_info['user'] = user
    return job_info

def get_oargrid_job_oar_jobs(oargrid_job_id = None, frontend_connection_params = None, timeout = False):
    """Return a list of tuples (oar job id, site), the list of individual oar jobs which make an oargrid job.

    :param oargrid_job_id: the oargrid job id.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use
      ``execo_g5k.config.g5k_configuration['default_timeout']``. None
      means no timeout.
    """
    if isinstance(timeout, bool) and timeout == False:
        timeout = g5k_configuration.get('default_timeout')
    process = get_process("oargridstat %i" % (oargrid_job_id,),
                          host = get_frontend_host(),
                          connection_params = make_connection_params(frontend_connection_params,
                                                                     default_frontend_connection_params))
    process.timeout = timeout
    process.pty = True
    process.run()
    if process.ok:
        job_specs = []
        for m in re.finditer("^\t(\w+) --> (\d+)", process.stdout, re.MULTILINE):
            site = m.group(1)
            if site not in get_g5k_sites():
                site = get_cluster_site(site)
            job_specs.append((int(m.group(2)), site))
        return job_specs
    else:
        raise ProcessesFailed([process])

def wait_oargrid_job_start(oargrid_job_id = None, frontend_connection_params = None, timeout = False):
    """Sleep until an oargrid job's start time.

    :param oargrid_job_id: the oargrid job id.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use
      ``execo_g5k.config.g5k_configuration['default_timeout']``. None
      means no timeout.
    """
    sleep(until = get_oargrid_job_info(oargrid_job_id, frontend_connection_params, timeout)['start_date'])

def get_oargrid_job_nodes(oargrid_job_id, frontend_connection_params = None, timeout = False):
    """Return an iterable of `execo.host.Host` containing the hosts of an oargrid job.

    :param oargrid_job_id: the oargrid job id.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use
      ``execo_g5k.config.g5k_configuration['default_timeout']``. None
      means no timeout.
    """
    if isinstance(timeout, bool) and timeout == False:
        timeout = g5k_configuration.get('default_timeout')
    process = get_process("oargridstat -wl %i 2>/dev/null || oargridstat -l %i 2>/dev/null" % (oargrid_job_id, oargrid_job_id),
                          host = get_frontend_host(),
                          connection_params = make_connection_params(frontend_connection_params,
                                                                     default_frontend_connection_params))
    process.timeout = timeout
    process.shell = process.pty = True
    process.run()
    if process.ok:
        host_addresses = re.findall("(\S+)", process.stdout, re.MULTILINE)
        return list(set([ Host(host_address) for host_address in host_addresses ]))
    else:
        raise ProcessesFailed([process])

def get_oargrid_job_key(oargrid_job_id = None, frontend_connection_params = None, timeout = False):
    """Return the filename of the oargrid job key

    :param oargrid_job_id: the oargrid job id.

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param timeout: timeout for retrieving. Default is False, which
      means use
      ``execo_g5k.config.g5k_configuration['default_timeout']``. None
      means no timeout.
    """
    return "/tmp/oargrid/oargrid_ssh_key_%s_%i" % (
        get_oargrid_job_info(
            oargrid_job_id,
            frontend_connection_params,
            timeout)['user'],
        oargrid_job_id)



import yaml, sys
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from scipy import stats

if __name__ == "__main__":
    fname = sys.argv[1]
    with open(fname, "r") as f:
        results = yaml.load(f, Loader=yaml.UnsafeLoader)
    arranged_results = {
        tcp_congestion_control:
            { num_flows:
                  [ result["bw"]
                    for result in results
                    if (result["params"]["num_flows"] == num_flows
                        and result["params"]["tcp_congestion_control"] == tcp_congestion_control) ]
              for num_flows in
              [ result["params"]["num_flows"] for result in results ] }
        for tcp_congestion_control in
        [ result["params"]["tcp_congestion_control"] for result in results ] }
    offset = -.18
    for tcp_congestion_control in arranged_results:
        plt.boxplot(list(arranged_results[tcp_congestion_control].values()),
                    positions = [ pos + offset for pos in arranged_results[tcp_congestion_control] ])
        plt.plot(
            sorted(arranged_results[tcp_congestion_control].keys()),
            [ stats.scoreatpercentile(arranged_results[tcp_congestion_control][num_flows], 50)
              for num_flows in sorted(arranged_results[tcp_congestion_control].keys())],
            label = tcp_congestion_control, linewidth=0.5)
        offset += .36
    plt.xlim(0, max(arranged_results[tcp_congestion_control]) + 1)
    #plt.xticks(arranged_results[tcp_congestion_control])
    plt.legend(loc = 'lower right')
    plt.xlabel('num flows')
    plt.ylabel('bandwith bits/s')
    plt.savefig("g5k_tcp_congestion.png")



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

""" A module based on `networkx <http://networkx.github.io/>`_ to create a
topological graph of the Grid'5000 platform. "Nodes" are used to represent
elements (compute nodes, switch, router, renater) and "Edges" are the network
links. Nodes has a kind data (+ power and core for compute nodes)
whereas edges has bandwidth and latency information.

All information comes from the Grid'5000 reference API.
"""

try:
    from networkx.drawing.nx_pydot import graphviz_layout
except:
    from networkx import graphviz_layout

from time import time
from execo import logger, Host
from execo.log import style
from execo.utils import do_once
from .oar import format_date
from .api_utils import get_g5k_sites, get_host_site, \
    get_host_cluster, get_g5k_clusters, get_cluster_hosts, \
    get_site_clusters, get_api_data, get_g5k_hosts, \
    get_network_equipment_attributes, get_host_shortname, \
    get_site_hosts
import networkx as nx
import re
from execo.utils import is_string

try:
    import matplotlib.pyplot as plt
except:
    pass


arbitrary_latency = 2.25E-3
suffix = '.grid5000.fr'

def _get_equip_name(equip, site):
    return "%s.%s" % (equip, site)

def _get_linecard_name(equip, site, lc_index):
    return "%s-lc%s.%s" % (equip, lc_index, site)

_parse_port_port_regex = re.compile('[^0-9]*([0-9]+)(:|/)([0-9]+)')
def _parse_port_port(p):
    mo = _parse_port_port_regex.match(p)
    if mo:
        target_lc = mo.group(1)
        target_port = mo.group(3)
        return target_lc, target_port
    else:
        return None, None

def _parse_port_uid(uid):
    return uid
    prefix, _, uid = uid.rpartition(' ')
    if prefix:
        logger.warn('uid %s prefixed with %s' % (uid, prefix))
    return uid

def _unique_link_key(*args):
    return '_'.join(sorted(args))

class g5k_graph(nx.MultiGraph):
    """Main graph representing the topology of the Grid'5000 platform. All
    nodes elements are defined with their FQDN"""

    def __init__(self, elements=None):
        """Create the :func:`~nx.MultiGraph` representing Grid'5000 network
        topology

        :param sites: add the topology of the given site(s)"""
        logger.debug("Initializing g5k_graph with elements %s" % (elements,))
        super(g5k_graph, self).__init__()
        self.data = get_api_data()
        self.graph['api_commit'] = self.data['network']['backbone'][0]['version']
        self.graph['date'] = format_date(time())

        if elements:
            if is_string(elements):
                elements = [elements]
            for e in elements:
                if e in get_g5k_sites():
                    self.add_site(e, self.data['sites'][e])
                elif e in get_g5k_clusters():
                    self.add_cluster(e, self.data['clusters'][e])
                else:
                    e = get_host_shortname(e)
                    if e in get_g5k_hosts():
                        self.add_host(e, self.data['hosts'][e])
            if len(self.get_sites()) > 1:
                self.add_backbone()

    # add/update/rm elements, public methods
    def add_host(self, host, data=None):
        """Add a host in the graph

        :param host: a string corresponding to the node name

        :param data: a dict containing the Grid'5000 host attributes"""
        if isinstance(host, Host):
            _host = get_host_shortname(host.address)
        else:
            _host = host
        if data:
            power = data['performance']['core_flops']
            cores = data['architecture']['nb_cores']
        else:
            power = 0
            cores = 0

        if len(self.get_host_adapters(_host)) == 0:
            logger.warning('Node %s has no valid network connection',
                           _host)
        logger.debug('Adding host %s', style.host(_host))
        self.add_node(_host, kind='node', power=power, cores=cores)
        for eq in self.get_host_adapters(_host):
            if eq['mounted']:
                self.add_equip(eq['switch'], get_host_site(_host))
                self._filter_equip_leaves()

    def rm_host(self, host):
        """Remove the host from the graph"""
        logger.debug('Removing host %s', style.host(host))
        self.remove_node(host)
        for eq in self.get_host_adapters(host):
            if not self._equip_has_nodes(eq['switch']):
                logger.debug('Removing equip %s', eq['switch'])
                self.rm_equip(eq['switch'])

    def add_cluster(self, cluster, data=None):
        """Add the cluster to the graph"""
        for h in get_cluster_hosts(cluster):
            self.add_host(h, self.data['hosts'][h])

    def rm_cluster(self, cluster):
        """Remove the cluster from the graph"""
        for h in get_cluster_hosts(cluster):
            self.rm_host(h)

    def add_site(self, site, data=None):
        """Add a site to the graph"""
        for h in get_site_hosts(site):
            self.add_host(h, self.data['hosts'][h])

    def rm_site(self, site):
        """Remove the site from the graph"""
        for h in get_site_hosts(site):
            self.rm_host(h)

    def _filter_equip_leaves(self):
        finished = False
        while not finished:
            finished = True
            for n in self.nodes():
                if self.nodes[n]['kind'] in ['switch', 'router', 'virtual']:
                    if len(list(self.neighbors(n))) < 1:
                        # changed this test from < 2 to < 1 without really understanding.
                        # with < 2 it removes sophia router
                        logger.debug('removing %s' % (n,))
                        self.remove_node(n)
                        finished = False
                        break
                    to_remove = True
                    for nb in self.neighbors(n):
                        if self.nodes[nb]['kind'] != 'linecard' or not nb.startswith(n + '-lc'):
                            to_remove = False
                        else:
                            if len(list(self.neighbors(nb))) >= 2:
                                to_remove = False
                    if to_remove:
                        removed = [ x for x in self.neighbors(n) if self.nodes[x]['kind'] == 'linecard' and x.startswith(n+'-lc') ]
                        removed.append(n)
                        for r in removed:
                            logger.debug('removing %s' % (r,))
                            self.remove_node(r)
                        finished = False
                        break

    def _checked_add_edge(self, node1, node2, key, **kwargs):
        if not self.has_edge(node1, node2, key):
            logger.debug('Adding link %s' % (key,))
            self.add_edge(node1, node2, key, **kwargs)

    def _checked_add_linecard(self, lc_node, backplane):
        if not self.has_node(lc_node):
            logger.debug('Adding linecard %s', lc_node)
            self.add_node(lc_node, kind='linecard', backplane=backplane)

    def _equip_uses_multiple_linecards(self, equip, site):
        num_linecards_with_ports = 0
        for lc in self.data['network'][site][equip]['linecards']:
            if 'ports' in lc:
                num_linecards_with_ports += 1
        if num_linecards_with_ports > 1:
            return True
        else:
            return False

    def _get_target_linecard_and_port_from_api(self, equip, linecard_index, port_index, site):
        # given the port <linecard_index>:<port_index> on equip/site
        # find (linecard_index, port_index) to which it is connected to on the target equipment
        # by searching in the target equipment network description
        # returns the tuple (linecard_index, port_index)
        port = self.data['network'][site][equip]['linecards'][linecard_index]['ports'][port_index]
        if 'kind' in port and port['kind'] == 'channel':
            uid = _parse_port_uid(self.data['network'][site][equip]['channels'][port['uid']]['uid'])
        else:
            uid = _parse_port_uid(port['uid'])
        if uid not in self.data['network'][site]:
            raise Exception('trying to find a linecard of equipment %s which is not in the network description of %s' % (uid, site))
        possible_targets = []
        for i_lc, lc in enumerate(self.data['network'][site][uid]['linecards']):
            if 'ports' in lc:
                for i_p, p in enumerate(lc['ports']):
                    if 'kind' in p and p['kind'] == 'channel':
                        if self.data['network'][site][uid]['channels'][p['uid']]['uid'] == equip:
                            possible_targets.append((i_lc, i_p))
                    else:
                        if 'uid' in p and _parse_port_uid(p['uid']) == equip:
                            possible_targets.append((i_lc, i_p))
        if 'kind' in port and port['kind'] == 'channel':
            if len(possible_targets) == 0:
                raise Exception
        targets = []
        if len(possible_targets) > 1:
            # need to disambiguate
            if 'kind' in port and port['kind'] == 'channel':
                # check that we have same number of ports on both sides, and all on the same linecards, else raise an exception
                chan_source = port['uid']
                num_lc_source = 0
                num_ports_source = 0
                this_port_index = None
                for i_lc, lc in enumerate(self.data['network'][site][equip]['linecards']):
                    lc_has_ports_on_channel = False
                    if 'ports' in lc:
                        for i_port, port in enumerate(lc['ports']):
                            if 'uid' in port and port['uid'] == chan_source:
                                if not lc_has_ports_on_channel:
                                    lc_has_ports_on_channel = True
                                    num_lc_source += 1
                                if linecard_index == i_lc and port_index == i_port:
                                    this_port_index = num_ports_source
                                num_ports_source += 1
                if num_lc_source > 1:
                    raise Exception("channel %s on %s has ports on %i linecards, we cannot disambiguate between ports if they are not given by g5k api" % (chan_source, equip, num_lc_source))
                if len(possible_targets) != num_ports_source:
                    raise Exception("channel %s on %s has %i ports but find %i corresponding ports on peer %i" % (chan_source, equip, num_ports_source, len(possible_targets), uid))
                target_linecards = set()
                for t in possible_targets:
                    target_linecards.add(t[0])
                if len(target_linecards) > 1:
                    raise Exception("channel %s on %s has a peer %s on which channel ports are on %i linecards, we cannot disambiguate between ports if they are not given by g5k api" % (chan_source, equip, uid, len(target_linecards)))
                # checks ok, just match ports in the order we get them
                targets.append(possible_targets[this_port_index])
            else:
                for target in possible_targets:
                    target_port_data = self.data['network'][site][uid]['linecards'][target[0]]['ports'][target[1]]
                    if 'port' in target_port_data:
                        target_lc, target_port = _parse_port_port(target_port_data['port'])
                        if target_lc is None or target_port is None:
                            logger.warn('unable to parse port spec %s of port %s:%s on %s of link from between %s(%s:%s %s)' % (target_port_data['port'], target[0], target[1], uid, equip, linecard_index, port_index, port))
                        if target_lc == linecard_index and target_port == port_index:
                            targets.append(target)
                    else:
                        logger.warn('no "port" entry in api network/%s/%s/linecards[%s]/ports[%s]' % (site, uid, target[0], target[1]))
        else:
            targets = possible_targets
        if len(targets) == 0:
            logger.warn('unable to find the target linecard on %s of link from %s(%s:%s %s)' % (uid, equip, linecard_index, port_index, port))
            if len(possible_targets) > 0:
                logger.warn('there are %s candidates %s, use the first possible one: %s' % (len(possible_targets), possible_targets, possible_targets[0]))
                return possible_targets[0]
            else:
                return (None, None)
        if len(targets) > 1:
            logger.warn('unable to disambiguate between multiple links to %s from %s(%s:%s %s). candidate linecards:ports are %s. Using the first possible one %s' % (uid, equip, linecard_index, port_index, port, targets, targets[0]))
            return targets[0]
        return targets[0]

    def _get_target_lc_and_port(self, equip, linecard_index, port_index, site):
        # given the port <linecard_index>:<port_index> on equip/site
        # find (linecard_index, port_index) to which is it connected to on the target equipment
        # by looking at the (optional) port specification and by searching in the target equipment network description, and comparing / complementing this informations
        # returns the tuple (linecard_index, port_index)
        port = self.data['network'][site][equip]['linecards'][linecard_index]['ports'][port_index]
        uid = _parse_port_uid(port['uid'])
        if 'port' in port:
            target_lc1, target_port1 = _parse_port_port(port['port'])
            if target_lc1 is None:
                logger.warn('unable to get from the port spec %s the target linecard of link between %s(%s:%s %s) and %s' % (port['port'], equip, linecard_index, port_index, port, uid))
            target_lc2, target_port2 = self._get_target_linecard_and_port_from_api(equip, linecard_index, port_index, site)
            if target_lc2 is None:
                logger.warn('unable to get from %s the target linecard of link between %s(%s:%s %s) and %s' % (uid, equip, linecard_index, port_index, port, uid))
            if (target_lc1 != target_lc2) or (target_port1 != target_port2):
                if not target_lc1 is None:
                    target_lc, target_port = target_lc1, target_port1
                else:
                    target_lc, target_port = target_lc2, target_port2
                logger.warn('mismatch between the linecards of link between %s(%s:%s %s) and %s: %s:%s vs %s:%s. Using the "less unlikely one" %s:%s' % (equip, linecard_index, port_index, port, uid, target_lc1, target_port1, target_lc2, target_port2, target_lc, target_port))
                return (target_lc, target_port)
        else:
            return self._get_target_linecard_and_port_from_api(equip, linecard_index, port_index, site)

    def _is_in_api(self, site, uid):
        if uid in self.data['hosts']:
            return True
        elif uid in self.data['network'][site]:
            return True
        elif 'renater' in uid:
            return True
        else:
            return False

    def _get_node_kind(self, site, uid):
        if uid in self.data['hosts']:
            return 'node'
        elif uid in self.data['network'][site]:
            return self.data['network'][site][uid]['kind']
        elif 'renater' in uid:
            #logger.warn('no kind found for %s. Guessing it is virtual only based on its name' % (uid,))
            return 'virtual'
        else:
            return None

    def add_equip(self, equip, site):
        """Add a network equipment """
        if equip not in self.data['network'][site]:
            logger.warn('Equipment %s not described in API' % (equip,))
            return
        data = self.data['network'][site][equip]
        equip_name = _get_equip_name(equip, site)
        if self.has_node(equip_name):
            recurse = False
        else:
            logger.debug('Adding equipment %s', equip_name)
            self.add_node(equip_name, kind=data['kind'],
                          backplane=data['backplane_bps'])
            recurse = True
        lc_data = data['linecards']
        multiple_linecards = self._equip_uses_multiple_linecards(equip, site)
        equip_bw = data['backplane_bps']
        for i_lc, lc in enumerate(lc_data):
            lc_node = _get_linecard_name(equip, site, i_lc)
            if 'ports' in lc:
                for i_port, port in enumerate(lc['ports']):
                    if 'uid' in port:
                        uid = _parse_port_uid(port['uid'])
                        kind = port.get('kind')
                        if kind == 'channel':
                            # if channel, simply replace the kind and uid by those of the channel
                            kind = data['channels'][uid]['kind']
                            uid = data['channels'][uid]['uid']
                        if not self._is_in_api(site, uid):
                            do_once((site, uid), logger.warn, 'unable to get kind of %s in %s, is it in g5k api?' % (uid, site))
                            continue
                        kind2 = self._get_node_kind(site, uid)
                        if not kind:
                            kind = kind2
                            if kind != 'node':
                                do_once((equip_name, i_lc, i_port), logger.warn, 'missing kind in port %s:%s %s of %s, using %s from %s' % (i_lc, i_port, port, equip_name, kind, uid))
                        elif not kind2:
                            logger.warn('missing kind in %s' % (uid,))
                        elif kind != kind2:
                            logger.warn('mismatching kind %s in port %s:%s %s of %s and kind %s from %s. Using %s' % (kind, i_lc, i_port, port, equip_name, kind2, uid, kind2))
                            kind = kind2
                        if not kind:
                            logger.error('unable to find kind of %s' % (uid, ))
                        port_bw = lc['rate'] if 'rate' not in port else port['rate']
                        if kind == 'virtual':
                            # in this situation, we don't know what
                            # kind is the target equipment, we need to
                            # discover it
                            if uid in self.data['network'][site]:
                                pass
                            elif uid in self.data['hosts']:
                                kind = 'virtual-node'
                                logger.warn('virtual link from %s(%s:%s %s) to node %s' % (equip_name, i_lc, i_port, port, uid))
                            else:
                                pass
                        target = None
                        if self.has_node(uid):
                            target = uid
                        if self.has_node(_get_equip_name(uid, site)):
                            target = _get_equip_name(uid, site)
                        if self.has_node(target):
                            if kind in ['node', 'virtual-node']:
                                for e in self.get_host_adapters(uid):
                                    if e['switch'] == equip:
                                        if multiple_linecards:
                                            self._checked_add_linecard(lc_node, lc.get('backplane_bps', data['backplane_bps']))
                                            self._checked_add_edge(equip_name, lc_node,
                                                                   _unique_link_key(equip_name, lc_node),
                                                                   bandwidth=equip_bw, active=True)
                                            self._checked_add_edge(lc_node, target,
                                                                   _unique_link_key(lc_node, target + '-' + e['device']),
                                                                   bandwidth=port_bw, active=e['mounted'])
                                        else:
                                            self._checked_add_edge(equip_name, target,
                                                                   _unique_link_key(equip_name, target + '-' + e['device']),
                                                                   bandwidth=min(port_bw, equip_bw), active=e['mounted'])
                            elif kind in ['switch', 'router'] and recurse:
                                if multiple_linecards:
                                    self._checked_add_linecard(lc_node, lc.get('backplane_bps', data['backplane_bps']))
                                    self._checked_add_edge(equip_name, lc_node,
                                                           _unique_link_key(equip_name, lc_node),
                                                           bandwidth=equip_bw, active=True)
                                    target_lc, target_port = self._get_target_lc_and_port(equip, i_lc, i_port, site)
                                    if not target_lc is None:
                                        if self._equip_uses_multiple_linecards(uid, site):
                                            self._checked_add_edge(lc_node, _get_linecard_name(uid, site, target_lc),
                                                                   _unique_link_key(lc_node, _get_linecard_name(uid, site, target_lc)),
                                                                   bandwidth=port_bw, active=True)
                                        else:
                                            self._checked_add_edge(lc_node, target,
                                                                   _unique_link_key(lc_node, target),
                                                                   bandwidth=port_bw, active=True)
                                    else:
                                        logger.error('unable to find the target linecard of link between %s(%s:%s %s) and %s. Skipping this link!' % (equip_name, i_lc, i_port, port, target))
                                else:
                                    target_lc, target_port = self._get_target_lc_and_port(equip, i_lc, i_port, site)
                                    if not target_lc is None:
                                        if self._equip_uses_multiple_linecards(uid, site):
                                            self._checked_add_edge(equip_name, _get_linecard_name(uid, site, target_lc),
                                                                   _unique_link_key(equip_name, _get_linecard_name(uid, site, target_lc)),
                                                                   bandwidth=min(port_bw, equip_bw), active=True)
                                        else:
                                            self._checked_add_edge(equip_name, target,
                                                                   _unique_link_key(equip_name, target),
                                                                   bandwidth=min(port_bw, equip_bw), active=True)
                                    else:
                                        logger.error('unable to find the target linecard of link between %s(%s:%s %s) and %s. Skipping this link!' % (equip_name, i_lc, i_port, port, target))
                        if 'renater' in uid:
                            # if uid != 'renater-' + site:
                            #     logger.error('renater node in %s has name %s which is not of the form renater-%s. Forcing to renater-%s' % (site, uid, site, site))
                            #     uid = 'renater-' + site
                            self.add_node(uid, kind='renater')
                            if multiple_linecards:
                                self._checked_add_linecard(lc_node, lc.get('backplane_bps', data['backplane_bps']))
                                self._checked_add_edge(equip_name, lc_node,
                                                       _unique_link_key(equip_name, lc_node),
                                                       bandwidth=equip_bw, active=True)
                                self._checked_add_edge(lc_node, uid,
                                                       _unique_link_key(lc_node, uid),
                                                       bandwidth=port_bw, active=True)
                            else:
                                self._checked_add_edge(equip_name, uid,
                                                       _unique_link_key(equip_name, uid),
                                                       bandwidth=min(port_bw, equip_bw), active=True)

                        elif kind in ['switch', 'router']:
                            if multiple_linecards:
                                self._checked_add_linecard(lc_node, lc.get('backplane_bps', data['backplane_bps']))
                                self._checked_add_edge(equip_name, lc_node,
                                                       _unique_link_key(equip_name, lc_node),
                                                       bandwidth=equip_bw, active=True)
                            if recurse:
                                self.add_equip(uid, site)

    def rm_equip(self, equip):
        """Remove an equipment from the node"""
        logger.debug('Removing equip %s', style.host(equip))
        self.remove_node(equip)
        if get_network_equipment_attributes(equip)['kind'] == 'router':
            lc_nodes = [x for x in self.nodes() if equip in x]
            logger.debug('Removing router linecard %s', ' '.join(lc_nodes))
            self.remove_nodes_from(lc_nodes)

    def add_backbone(self):
        """Add the nodes corresponding to Renater equipments"""
        logger.debug('Add %s network', style.emph('Renater'))
        backbone = self.data['network']['backbone']
        for equip in backbone:
            src = equip['uid']
            self.add_node(src, kind='renater')
            for lc in equip['linecards']:
                for port in lc['ports']:
                    if 'uid' in port and 'renater-' in _parse_port_uid(port['uid']):
                        port_bw = lc['rate'] if 'rate' not in port else port['rate']
                        latency = port['latency'] if 'latency' in port \
                            else arbitrary_latency
                        kind = 'renater' if 'kind' not in port else port['kind']
                        dst = _parse_port_uid(port['uid'])
                        logger.debug('* %s (%s, bw=%s, lat=%s)', dst, kind,
                                      port_bw, latency)
                        self.add_node(dst, kind=kind)
                        if not self.has_edge(src, dst):
                            self.add_edge(src, dst,
                                          _unique_link_key(src, dst),
                                          bandwidth=port_bw,
                                          latency=latency, active=True)
        # Removing unused one
        logger.debug('Removing unused Renater equipments')
        used_elements = []
        for site in self.get_sites():
            dests = [ s for s in self.get_sites() if s != site ]
            for dest in dests:
                gw_src = self.get_site_router(site)[0]
                gw_dst = self.get_site_router(dest)[0]
                if not gw_src is None and not gw_dst is None:
                    for element in [el for el in nx.shortest_path(self, gw_src, gw_dst) if 'renater' in el]:
                        if element not in used_elements:
                            used_elements.append(element)

        for element, _ in [n for n in self.nodes(data=True) if n[1]['kind'] == 'renater']:
            if element not in used_elements:
                logger.debug('removing %s' % (element,))
                self.remove_node(element)

    def rm_backbone(self):
        """Remove all elements from the backbone"""
        self.remove_nodes_from(self.get_backbone())

    # get elements, public methods
    def get_hosts(self, cluster=None, site=None):
        """Return the list of nodes corresponding to hosts"""
        if cluster:
            return [x for x in self.nodes(True) if cluster in x[0] and
                          x[1]['kind'] == 'node']
        elif site:
            return [x for x in self.nodes(True) if site == get_host_site(x[0]) and
                          x[1]['kind'] == 'node']
        else:
            return [x for x in self.nodes(True) if x[1]['kind'] == 'node']

    def get_clusters(self, site=None):
        """Return the list of clusters"""
        if site:
            return list(set([get_host_cluster(x[0]) for x in self.nodes(True) if get_host_site(x[0]) and x[1]['kind'] == 'node']))
        else:
            return list(set([get_host_cluster(x[0]) for x in self.nodes(True) if x[1]['kind'] == 'node']))

    def get_host_neighbours(self, host):
        """Return the compute nodes that are connected to the same switch,
        router or linecard"""

        switch = self.get_host_adapters(host)[0]['switch']
        return [x for x in self.get_equip_hosts(switch) if x != host]

    def get_equip_hosts(self, equip):
        """Return the nodes which are connected to the equipment"""
        hosts = []
        if self.nodes[equip]['kind'] == 'router':
            lcs = self.neighbors(equip)
        else:
            lcs = ['equip']
        for lc in lcs:
            for n in self.neighbors(lc):
                if self.nodes[n]['kind'] == 'node':
                    hosts.append(n)

        return hosts

    def get_site_router(self, site):
        """Return the node corresponding to the router of a site"""
        site_routers = [n for n in self.nodes(True) if n[1]['kind'] == 'router' and site in n[0]]
        if len(site_routers) > 0:
            return site_routers[0]
        else:
            return (None, {})

    def get_sites(self):
        """Return the list of sites"""
        return list(set([get_host_site(x[0]) for x in self.nodes(True) if x[1]['kind'] == 'node']))

    def get_backbone(self):
        """Return """
        return [x for x in self.nodes(True) if x[1]['kind'] == 'renater']

    def get_host_adapters(self, host):
        """Return the mountable network interfaces from a host"""
        try:
            if host in self.data['hosts']:
                return [m for m in self.data['hosts'][host]['network_adapters']
                        if 'switch' in m
                        and not m['management']
                        and m['mountable']
                        and m['switch']
                        and m['interface'] == 'Ethernet']
        except:
            logger.warning("Wrong description for host %s" % style.host(host))
            logger.debug("host's network_adapters = %s" % (self.data['hosts'][host]['network_adapters'],))
            return []

    def _equip_has_nodes(self, equip):
        """ """
        data = get_network_equipment_attributes(equip)
        if data['kind'] == 'router':
            return True
        for lc in [n for n in data['linecards'] if 'ports' in n]:
            for port in sorted([p for p in lc['ports'] if 'uid' in p]):
                kind = port['kind'] if 'kind' in port else lc['kind']
                if kind == 'node' and self.has_node(_parse_port_uid(port['uid'])):
                    return True
        return False


def treemap(gr, nodes_legend=None, edges_legend=None, nodes_labels=None,
            layout='neato', compact=False):
    """Create a treemap of the topology and return a matplotlib figure

    :param nodes_legend: a dict of dicts containing the parameter used to draw
     the nodes, such as 'myelement': {'color': '#9CF7BC', 'shape': 'p',
     'size': 200}

    :param edges_legend: a dict of dicts containing the parameter used to draw
     the edges, such as bandwidth: {'width': 0.2, 'color': '#666666'}

    :param nodes_labels: a dict of dicts containing the font parameters for
     the labels, such as 'myelement ': {'nodes': {}, 'font_size': 8,
     'font_weight': 'bold', 'str_func': lambda n: n.split('.')[1].title()}

    :param layout: the graphviz tool to be used to compute node position

    :param compact: represent only on node for a cluster/cabinet

    WARNING: This function use matplotlib.figure that by default requires a
    DISPLAY. If you want use this on a headless host, you need to change the
    matplotlib backend before to import execo_g5k.topology module.
    """

    base_size = 2

    _default_color = '#000000'
    _default_shape = 'o'
    _default_size = 100
    _default_width = 0.8
    _default_font_size = 10
    _default_font_weight = 'normal'

    def _default_str_func(n):
        return n.split('.')[0]

    def _default_nodes_legend():
        """Create a default legend for the nodes"""
        return {'renater':
                {'color': '#9CF7BC', 'shape': 'p', 'size': 200},
                'router':
                {'color': '#BFDFF2', 'shape': '8', 'size': 300,
                 'width': 0.5},
                'switch':
                {'color': '#F5C9CD', 'shape': 's', 'size': 100,
                 'width': 0.2},
                'host':
                {'color': '#F0F7BE', 'shape': 'o', 'size': 30,
                 'width': 0.2},
                'cluster':
                {'color': '#F0F7BE', 'shape': 'd', 'size': 200,
                 'width': _default_width},
                'default':
                {'color': _default_color, 'shape': _default_shape,
                 'size': _default_size},
                'linecard':
                {'size': 10, 'shape': '^', 'color': 'w', 'width': 0.1},
                }

    def _default_edges_legend():
        """Defines the width and color of the edges based on bandwidth"""
        return {100000000: {'width': 0.2, 'color': '#666666'},
                1000000000: {'width': 0.4, 'color': '#666666'},
                3000000000: {'width': 0.6, 'color': '#333333'},
                10000000000: {'width': 1.0, 'color': '#111111'},
                20000000000: {'width': 2.0, 'color': '#111111'},
                30000000000: {'width': 3.0, 'color': '#111111'},
                40000000000: {'width': 4.0, 'color': '#111111'},
                'default': {'width': _default_width, 'color': _default_color}}

    def _default_nodes_labels(compact=False):
        """Defines the font labels"""

        def _default_str_func(n):
            return n.split('.')[0]

        return {'renater':
                {'nodes': {},
                 'font_size': base_size * 4,
                 'font_weight': 'normal',
                 'str_func': _default_str_func},
                'router':
                {'nodes': {},
                 'font_size': base_size * 4,
                 'font_weight': 'bold'},
                'switch':
                {'nodes': {},
                 'font_size': base_size * 4,
                 'font_weight': 'normal'},
                'cluster':
                {'nodes': {},
                 'font_size': base_size * 4,
                 'font_weight': 'normal'},
                'host':
                {'nodes': {},
                 'font_size': base_size * 3,
                 'font_weight': 'normal'},
                'default':
                {'nodes': {},
                 'font_size': _default_font_size,
                 'font_weight': _default_font_weight,
                 'str_func': _default_str_func},
                'linecard':
                {'nodes': {},
                 'font_size': base_size * 3,
                 'str_func': _default_str_func}
                }

    grdot = gr.copy()

    # Setting legend and labels
    _nodes_legend = _default_nodes_legend()
    _edges_legend = _default_edges_legend()
    _nodes_labels = _default_nodes_labels(compact)
    if nodes_legend:
        _nodes_legend.update(nodes_legend)
    if edges_legend:
        _edges_legend.update(edges_legend)
    if nodes_labels:
        _nodes_labels.update(nodes_labels)

    if not compact:
        elements = ['renater', 'router', 'switch', 'node', 'linecard']
    else:
        for site in grdot.get_sites():
            for cluster, data in grdot.get_clusters().items():
                for equip, radicals in data['equips'].items():
                    grdot.add_node(cluster + '\n' + radicals,
                                {'kind': 'cluster'})
                    grdot.add_edge(cluster + '\n' + radicals, equip,
                                {'bandwidth': data['bandwidth']})

        grdot.remove_nodes_from([n[0] for n in grdot.nodes(True) if n[1]['kind'] == 'node'])

        elements = ['renater', 'router', 'switch', 'cluster']

    logger.debug('Legend and labels initialized')

    # substitute kind host to kind node in all the graph, as node is a reserved dot word
    for n in grdot.nodes(True):
        if n[1].get('kind') == 'node':
            n[1]['kind'] = 'host'

    # Initializing plot
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111)

    logger.debug('Defining positions')
    try:
        pos = graphviz_layout(grdot, prog=layout)
    except:
        logger.warning('Error in generating graphviz layout, will use ' +
                       'spring layout that does not scale well ...')
        raise
        pos = nx.spring_layout(grdot, iterations=100)
    # Adding the nodes
    for k in elements:
        nodes = [node[0] for node in grdot.nodes(data=True)
                 if 'kind' in node[1] and node[1]['kind'] == k]
        if k not in _nodes_legend:
            _nodes_legend[k] = _nodes_legend['default']
        nodes = nx.draw_networkx_nodes(grdot, pos, nodelist=nodes,
                                       node_shape=_nodes_legend[k]['shape']
                                       if 'shape' in _nodes_legend[k] else
                                       _default_shape,
                                       node_color=_nodes_legend[k]['color']
                                       if 'color' in _nodes_legend[k] else
                                       _default_color,
                                       node_size=_nodes_legend[k]['size']
                                       if 'size' in _nodes_legend[k] else
                                       _default_size,
                                       linewidths=_nodes_legend[k]['width']
                                       if 'width' in _nodes_legend[k] else
                                       _default_width)

    # Adding the edges
    for bandwidth, params in _edges_legend.items():
        if bandwidth != 'other':
            edges = [(edge[0], edge[1]) for edge in grdot.edges(data=True)
                     if 'bandwidth' in edge[2] and edge[2]['bandwidth'] == bandwidth]
            nx.draw_networkx_edges(grdot, pos, edgelist=edges,
                                   width=params['width'] if 'width' in params
                                   else _default_width,
                                   edge_color=params['color'] if 'color' in params
                                   else _default_color)
    edges = [(edge[0], edge[1]) for edge in grdot.edges(data=True)
             if edge[2]['bandwidth'] not in _edges_legend]

    nx.draw_networkx_edges(grdot, pos, edgelist=edges,
                           width=_edges_legend['default']['width'],
                           edge_color=_edges_legend['default']['color'])
    # Adding the labels
    for node, data in grdot.nodes(data=True):
        if 'nodes' not in _nodes_labels[data['kind']]:
            _nodes_labels[data['kind']]['nodes'] = {}
        if data['kind'] in _nodes_labels:
            _nodes_labels[data['kind']]['nodes'][node] = _nodes_labels[data['kind']]['str_func'](node) \
                if 'str_func' in _nodes_labels[data['kind']] else _default_str_func(node)
        else:
            _nodes_labels['default']['nodes'][node] = _nodes_labels['default']['str_func'](node)

    for data in _nodes_labels.values():
        nx.draw_networkx_labels(grdot, pos, labels=data['nodes'],
                                font_size=data['font_size']
                                if 'font_size' in data else _default_font_size,
                                font_weight=data['font_weight']
                                if 'font_weight' in data else _default_font_weight)

    plt.axis('off')
    plt.tight_layout()

    title = 'Created by execo_g5k.topology \n%s\nAPI commit %s' % \
        (grdot.graph['date'], grdot.graph['api_commit'])
    plt.text(0.1, 0, title, transform=ax.transAxes)

    return fig




#!/usr/bin/env python

# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
# This file is part of Execo, released under the GNU Lesser Public
# License, version 3 or later.

from __future__ import print_function
from distutils.command.clean import clean as _clean
from distutils.dir_util import remove_tree
from distutils import log
import sys, subprocess, os, textwrap, shutil, re
import codecs

try:
    from setuptools import setup
    from setuptools.command.install import install as _install
    from setuptools.command.build_py import build_py as _build_py
    from setuptools.command.sdist import sdist as _sdist
except:
    from distutils.core import setup
    from distutils.command.install import install as _install
    from distutils.command.build_py import build_py as _build_py
    from distutils.command.sdist import sdist as _sdist

try:
    from sphinx.setup_command import BuildDoc
except:
    pass

def get_git_version():
    # returns git tag / sha as string
    # returns None if not available
    try:
        p = subprocess.Popen(["git", "describe", "--tags", "--dirty", "--always"],
                             stdout = subprocess.PIPE, stderr = subprocess.PIPE,
                             universal_newlines = True)
    except EnvironmentError:
        return None
    version = p.communicate()[0].rstrip()
    if p.returncode != 0:
        return None
    return version

VERSION_PY = """
# Do not edit. This file is originally generated from git information.
# Distribution tarballs contain a pre-generated copy of this file.
__version__ = '%s'
"""

def update_version_py():
    version_filename = "src/execo/_version.py"
    version = get_git_version()
    if version or not os.path.isfile(version_filename):
        if not version: version = "UNKNOWN"
        with open(version_filename, "w") as f:
            f.write(VERSION_PY % version)

def get_version():
    gitversion = get_git_version()
    if gitversion:
        return gitversion
    try:
        with open("src/execo/_version.py") as f:
            for line in f.readlines():
                mo = re.match("__version__ = '([^']*)'", line)
                if mo:
                    version = mo.group(1)
                    return version
    except EnvironmentError:
        pass
    return "UNKNOWN"

def read_file(filename, start_marker = None, end_marker = None):
    s = ""
    with codecs.open(filename, mode='r', encoding='utf-8') as ifh:
        insection = (start_marker == None)
        for line in ifh:
            line = line.rstrip()
            if start_marker != None and line == start_marker and not insection:
                insection = True
                continue
            if end_marker != None and line == end_marker and insection:
                insection = False
                break
            if insection:
                s += line + "\n"
    return s.rstrip()

def extract_conf(fh, source_file, marker):
    s = read_file(source_file, "# _STARTOF_ " + marker, "# _ENDOF_ " + marker)
    s = textwrap.dedent(s)
    for l in s.splitlines():
        print("# " + l, file=fh)
    print("\n", file=fh)

def generate_conf_template(datadir):
    try:
        os.makedirs(os.path.join(datadir, "share", "execo"))
    except os.error:
        pass
    with open(os.path.join(datadir, "share", "execo", "execo.conf.py.sample"), "w") as fh:
        print("# sample execo user configuration", file=fh)
        print("# copy this file to ~/.execo.conf.py and edit/modify it appropriately", file=fh)
        print(file=fh)
        print("# import logging, os, sys", file=fh)
        print(file=fh)
        extract_conf(fh, os.path.join("src", "execo", "config.py"), "configuration")
        extract_conf(fh, os.path.join("src", "execo", "config.py"), "default_connection_params")
        extract_conf(fh, os.path.join("src", "execo_g5k", "config.py"), "g5k_configuration")
        extract_conf(fh, os.path.join("src", "execo_g5k", "config.py"), "default_frontend_connection_params")
        extract_conf(fh, os.path.join("src", "execo_g5k", "config.py"), "default_oarsh_oarcp_params")

def copy_additional_files(datadir):
    try:
        os.makedirs(os.path.join(datadir, "share", "execo"))
    except os.error:
        pass

class build_py(_build_py):
    def run(self):
        update_version_py()
        _build_py.run(self)

class sdist(_sdist):
    def run(self):
        update_version_py()
        _sdist.run(self)

class install(_install):
    def run(self):
        _install.run(self)
        self.execute(generate_conf_template, (self.install_data,),
                     msg = "Generate execo configuration template")
        self.execute(copy_additional_files, (self.install_data,),
                     msg = "Copying additional files")

class install_doc(_install):
    def run(self):
        #self.run_command('build_doc')
        build = self.get_finalized_command('build')
        build_dir = os.path.join(os.path.abspath(build.build_base), "sphinx", "html")
        build_dir = os.path.abspath(build_dir)
        self.copy_tree(build_dir, os.path.join(self.install_data, "share", "doc", "execo", "html"))

class clean(_clean):
    def run(self):
        if self.all:
            sphinx_dir = os.path.join(self.build_base, "sphinx")
            if os.path.exists(sphinx_dir):
                remove_tree(sphinx_dir, dry_run = self.dry_run)
            else:
                log.warn("'%s' does not exist -- can't clean it",
                         sphinx_dir)
            def rm(f):
                try:
                    os.unlink(f)
                    log.info("removing " + f)
                except:
                    log.warn("can't clean " + f),
            rm("MANIFEST")
            rm("src/execo/_version.py")
        _clean.run(self)

if __name__ == "__main__":

    try:
        cmdclass = { 'build_py': build_py,
                     'sdist': sdist,
                     'install': install,
                     'build_doc': BuildDoc,
                     'install_doc': install_doc,
                     'clean': clean }
    except:
        cmdclass = { 'build_py': build_py,
                     'sdist': sdist,
                     'install': install,
                     'install_doc': install_doc,
                     'clean': clean }

    name = 'execo'
    version = get_version()

    setup(cmdclass = cmdclass,
          name = name,
          license = 'GNU GPL v3',
          version = version,
          description = "Python API for asynchronous control of local or remote, standalone or parallel, unix processes.",
          long_description = read_file("README.md"),
          author = 'Matthieu Imbert',
          author_email = 'matthieu.imbert@inria.fr',
          url = 'https://gitlab.inria.fr/mimbert/execo',
          package_dir = {'': 'src'},
          packages = [ 'execo', 'execo_g5k', 'execo_engine' ],
          package_data = { 'execo': ['execo-chainput'] },
          classifiers = [ 'Development Status :: 4 - Beta',
                          'Environment :: Console',
                          'Intended Audience :: Developers',
                          'Intended Audience :: Information Technology',
                          'Intended Audience :: Science/Research',
                          'Intended Audience :: System Administrators',
                          'Operating System :: POSIX :: Linux',
                          'Programming Language :: Python :: 2.6',
                          'Programming Language :: Python :: 2.7',
                          'Programming Language :: Python :: 3.8',
                          'Topic :: Software Development',
                          'Topic :: System :: Clustering',
                          'Topic :: System :: Distributed Computing'],
          platforms = [ 'unix' ]
          )



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

import threading, os, fcntl, math, sys
import pickle as pickle
if sys.version_info >= (3,):
    import pickle
else:
    import cPickle as pickle
from .log import logger

def geom(range_min, range_max, num_steps):
    """Return a geometric progression from range_min to range_max with num_steps"""
    if num_steps == 0: return []
    if num_steps == 1: return [float(range_min)]
    return ([ float(range_min) ]
            + [ float(range_min) * math.pow(math.pow(float(range_max) / float(range_min), 1.0 / (int(num_steps) - 1)), i)
                for i in range(0, int(num_steps)) ][1:-1]
            + [ float(range_max) ])

def igeom(range_min, range_max, num_steps):
    """Return an integer geometric progression from range_min to range_max with num_steps"""
    if num_steps == 0: return []
    if num_steps == 1: return [int(range_min)]
    return sorted(set([ int(range_min) ]
                      + [ int(round(x)) for x in geom(range_min, range_max, num_steps) ]
                      + [ int(range_max) ]))

class HashableDict(dict):

    """Hashable dictionnary. Beware: must not mutate it after its first use as a key."""

    def __key(self):
        return tuple((k,self[k]) for k in sorted(self))

    def __hash__(self):
        return hash(self.__key())

def sweep(parameters):

    """Generates all combinations of parameters.

    The aim of this function is, given a list of experiment parameters
    (named factors), and for each parameter (factor), the list of
    their possible values (named levels), to generate the cartesian
    product of all parameter values, for a full factorial experimental
    design (*The Art Of Computer Systems Performance Analysis,
    R. Jain, Wiley 1991*).

    More formally: given a a dict associating factors as keys and the
    list of their possible levels as values, this function will return
    a list of dict corresponding to all cartesian product of all level
    combinations. Each dict in the returned list associates the
    factors as keys, and one of its possible levels as value.

    In the given factors dict, if for a factor X (key), the value
    associated is a dict instead of a list of levels, then it will use
    the keys of the sub-dict as levels for the factor X, and the
    values of the sub-dict must also be some dict for factor / levels
    combinations which will be explored only for the corresponding
    levels of factor X. This is kind of recursive sub-sweep and it
    allows to explore some factor / level combinations only for some
    levels of a given factor.

    The returned list contains `execo_engine.sweep.HashableDict`
    instead of dict, which is a simple subclass of dict, so that
    parameters combinations can be used as dict keys (but don't modify
    them in such cases)

    Examples:

    >>> sweep({
    ...     "param 1": ["a", "b"],
    ...     "param 2": [1, 2]
    ...     })
    [{'param 1': 'a', 'param 2': 1}, {'param 1': 'a', 'param 2': 2}, {'param 1': 'b', 'param 2': 1}, {'param 1': 'b', 'param 2': 2}]

    >>> sweep({
    ...     "param 1": ["a", "b"],
    ...     "param 2": {
    ...         1: {
    ...             "param 1 1": [ "x", "y" ],
    ...             "param 1 2": [ 0.0, 1.0 ]
    ...             },
    ...         2: {
    ...             "param 2 1": [ -10, 10 ]
    ...             }
    ...         }
    ...     })
    [{'param 1 2': 0.0, 'param 1 1': 'x', 'param 1': 'a', 'param 2': 1}, {'param 1 2': 0.0, 'param 1 1': 'y', 'param 1': 'a', 'param 2': 1}, {'param 1 2': 1.0, 'param 1 1': 'x', 'param 1': 'a', 'param 2': 1}, {'param 1 2': 1.0, 'param 1 1': 'y', 'param 1': 'a', 'param 2': 1}, {'param 2 1': -10, 'param 1': 'a', 'param 2': 2}, {'param 2 1': 10, 'param 1': 'a', 'param 2': 2}, {'param 1 2': 0.0, 'param 1 1': 'x', 'param 1': 'b', 'param 2': 1}, {'param 1 2': 0.0, 'param 1 1': 'y', 'param 1': 'b', 'param 2': 1}, {'param 1 2': 1.0, 'param 1 1': 'x', 'param 1': 'b', 'param 2': 1}, {'param 1 2': 1.0, 'param 1 1': 'y', 'param 1': 'b', 'param 2': 1}, {'param 2 1': -10, 'param 1': 'b', 'param 2': 2}, {'param 2 1': 10, 'param 1': 'b', 'param 2': 2}]
    """

    result = [HashableDict()]
    for key, val in parameters.items():
        if len(val) == 0: continue
        newresult = []
        for i in result:
            if isinstance(val, dict):
                for subkey, subval in val.items():
                    for subcombs in sweep(subval):
                        subresult = HashableDict(i)
                        subresult.update({key: subkey})
                        subresult.update(subcombs)
                        newresult.append(subresult)
            else:
                for j in val:
                    subresult = HashableDict(i)
                    subresult.update({key: j})
                    newresult.append(subresult)
        result = newresult
    return result

# context manager for opening and locking files
# beware: for locking purpose, the file is always opened in mode "ab+"
# which is the only mode allowing both locking the file and having
# read write access to it. but it forces to handle correctly file
# position and truncation
class _openlock():

    def __init__(self, filename):
        self.__filename = filename

    def __enter__(self):
        self.__file = open(self.__filename, "ab+")
        fcntl.lockf(self.__file, fcntl.LOCK_EX)
        return self.__file

    def __exit__(self, t, v, traceback):
        self.__file.flush()
        os.fsync(self.__file.fileno())
        fcntl.lockf(self.__file, fcntl.LOCK_UN)
        self.__file.close()
        return False

class ParamSweeper(object):

    """Multi-process-safe, thread-safe and persistent iterable container to iterate over a list of experiment parameters (or whatever, actually).

    The aim of this class is to provide a convenient way to iterate
    over several experiment configurations (or anything else). It is
    an iterable container with the following characteristics:

    - each element of the iterable has four states:

      - *todo*

      - *inprogress*

      - *done*

      - *skipped*

    - at beginning, each element is in state *todo*

    - client code can mark any element *done* or *skipped*

    - when iterating over it, you always get the next item in *todo*
      state

    - this container has automatic persistence of the element states
      *done* and *inprogress* (but not state *skipped*) to disk: If
      later you instanciate a container with the same persistence
      directory (path to is given to constructor), then the elements
      of the container will be taken from the constructor argument,
      but states *done* or *inprogress* will be loaded from persistent
      state.

    - this container is thread-safe and multi-process-safe. Multiple
      threads can concurrently use a single ParamSweeper
      object. Multiple threads or processes on the same or different
      hosts can concurrently use several ParamSweeper instances
      sharing the same persistence directory. With sufficiently recent
      linux kernels and nfs servers / clients, it will work on a
      shared nfs storage (current implementation uses python flock,
      which should work since kernel 2.6.12. see
      http://nfs.sourceforge.net/#faq_d10). All threads sharing a
      ParamSweeper instance synchronize through in-process locks, and
      all threads / processes with different instances of ParamSweeper
      sharing the same persistent directory synchronize through the
      persisted state.

    This container is intended to be used in the following way: at the
    beginning of the experiment, you initialize a ParamSweeper with
    the list of experiment configurations (which can result from a
    call to `execo_engine.sweep.sweep`, but not necessarily) and a
    directory for the persistence. During execution, you request
    (possibly from several concurrent threads or processes) new
    experiment configurations with
    `execo_engine.sweep.ParamSweeper.get_next`, mark them *done* or
    *skipped* with `execo_engine.sweep.ParamSweeper.done` and
    `execo_engine.sweep.ParamSweeper.skip`. At a later date, you can
    relaunch the same script, it will continue from where it left,
    also retrying the skipped configurations. This works well when
    used with `execo_engine.engine.Engine` startup option ``-c``
    (continue experiment in a given directory).

    `execo_engine.sweep.ParamSweeper.skip` is intended to be used when
    you got a combination with
    `execo_engine.sweep.ParamSweeper.get_next` but the processing of
    this combination has failed and you don't want to retry it later
    in the same run of your script.

    But the skipped state is not written to disk, if you relaunch your
    script later, the previously skipped combinations will again be
    iterated over.

    The intent is for situations where, for example, some combinations
    cannot be processed because, for example, there are not enough
    available resources to process them. You may then want to relaunch
    your script later (after making sure more resources are available)
    to process these skipped combinations.

    If there are some combinations that you never want to retry, even
    in subsequent runs of your script, then mark them as done with
    `execo_engine.sweep.ParamSweeper.done`.

    If there is a combination that you failed to process but you want
    to retry it in the current run of your script, mark it as canceled
    with `execo_engine.sweep.ParamSweeper.cancel`.

    State *inprogress* is stored on disk to avoid concurrent processes
    to get the same elements from different ParamSweeper instances (to
    avoid duplicating work). In some cases (for example if a process
    has crashed without marking an element *done* or *skipped* or
    canceling it), you may want to reset the *inprogress* state. This
    can be done by removing the file ``inprogress`` in the persistence
    directory (this can even be done while some ParamSweeper are
    instanciated and using it).

    ParamSweeper handle crashes in the following ways: if it crashes
    (or is killed) while synchronizing to disk, in the worst case, the
    current element marked done can be lost (i.e. other ParamSweepers
    or later instanciations will not see it marked done), or the whole
    list of inprogress elements can be lost.

    The ParamSweeper code assumes that in typical usage, there may be
    a huge number of elements to iterate on (and accordingly, the list
    of done elements will grow huge too), and that the number of
    inprogress elements will stay reasonably low. The whole iterable
    of elements is (optionally) written to disk only once, at
    ParamSweeper construction. The set of done elements can only grow
    and is incrementaly appended. The set of inprogress elements is
    fully read from and written to disk at each operation, thus this
    may become a bottleneck to ParamSweeper performance if the set of
    inprogress elements is big.

    """

    def __init__(self, persistence_dir, sweeps = None, save_sweeps = False, name = None):
        """
        :param persistence_dir: path to persistence directory. In this
          directory will be created to python pickle files: ``done``
          and ``inprogress`` This files can be erased if needed.

        :param sweeps: An iterable, what to iterate on. If None
          (default), try to load it from ``persistence_dir``

        :param save_sweeps: boolean. default False. If True, the
          sweeps are written to disk during initialization (this may
          take some time but occurs only once)

        :param name: a convenient name to identify an instance in
          logs. If None, compute one from persistence_dir.
        """
        self.__lock = threading.RLock()
        self.__persistence_dir = persistence_dir
        try:
            os.makedirs(self.__persistence_dir)
        except os.error:
            pass
        self.__name = name
        if not self.__name:
            self.__name = os.path.basename(self.__persistence_dir)

        self.__done = set()
        self.__inprogress = set()
        self.__skipped = set()

        self.__filtered_done = set()
        self.__filtered_inprogress = set()
        self.__filtered_skipped = set()
        # __filtered_done, __filtered_inprogress, __filtered_skipped
        # are the intersections of __sweeps and __done, __inprogress,
        # __skipped. They exist because:
        #
        # - client may call set_sweeps with different sweeps, still we
        #   want to remember the complete list of __done,
        #   __inprogress, __skipped
        #
        # - different ParamSweeper instances may share the same
        #   storage though having different individual
        #   __sweeps. __inprogress and __done on storage will be the
        #   union of all inprogress and done, so the ParamSweeper must
        #   be prepared to deal correctly with __inprogress and __done
        #   containing elements not in *its* __sweeps (and it must not
        #   discard them)
        #
        # - on the other hand, when displaying with __str__(),
        #   stats(), or when retrieving with get_done(),
        #   get_inprogress(), get_skipped(), client expects to get
        #   only lists of done, inprogress, skipped relative to the
        #   current __sweeps.
        #
        # - we could filter (do the intersection with __sweeps) only
        #   when displaying or returning the lists, but it's a costly
        #   operation to do the full intersection, whereas doing it
        #   incrementaly is fast.

        self.__remaining = set()
        self.__done_filepos = None

        self.set_sweeps(sweeps, save_sweeps)

    def set_sweeps(self, sweeps = None, save_sweeps = False):
        """Change the list of what to iterate on.

        :param sweeps: iterable

        :param save_sweeps: boolean. default False. If True, the
          sweeps are written to disk.
        """
        with self.__lock:
            if sweeps:
                self.__sweeps = set(sweeps)
                if save_sweeps:
                    with _openlock(os.path.join(self.__persistence_dir, "sweeps")) as sweeps_file:
                        sweeps_file.truncate(0)
                        pickle.dump(self.__sweeps, sweeps_file)
            else:
                with _openlock(os.path.join(self.__persistence_dir, "sweeps")) as sweeps_file:
                    sweeps_file.seek(0, os.SEEK_SET)
                    self.__sweeps = pickle.load(sweeps_file)
            self.full_update()

    def __nolock_full_update(self, done_file, inprogress_file):
        self.__done.clear()
        self.__done_filepos = 0
        done_file.seek(0, os.SEEK_SET)
        while True:
            try:
                self.__done.add(pickle.load(done_file))
                self.__done_filepos = done_file.tell()
            except:
                done_file.truncate(self.__done_filepos)
                break
        inprogress_file.seek(0, os.SEEK_SET)
        try:
            self.__inprogress = pickle.load(inprogress_file)
        except:
            inprogress_file.truncate(0)
            self.__inprogress.clear()
        self.__remaining = set(self.__sweeps).difference(self.__done, self.__skipped, self.__inprogress)
        self.__filtered_done = self.__done.intersection(self.__sweeps)
        self.__filtered_inprogress = self.__inprogress.intersection(self.__sweeps)
        self.__filtered_skipped = self.__skipped.intersection(self.__sweeps)

    def full_update(self):
        """Reload completely the ParamSweeper state from disk (may take some time)."""
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_full_update(done_file, inprogress_file)

    def __nolock_update(self, done_file, inprogress_file):
        done_file.seek(0, os.SEEK_END)
        new_done_filepos = done_file.tell()
        if new_done_filepos >= self.__done_filepos:
            if new_done_filepos > self.__done_filepos:
                done_file.seek(self.__done_filepos, os.SEEK_SET)
                new_done = list()
                while True:
                    try:
                        new_done.append(pickle.load(done_file))
                        self.__done_filepos = done_file.tell()
                    except:
                        done_file.truncate(self.__done_filepos)
                        break
                self.__done.update(new_done)
                self.__remaining.difference_update(new_done)
                self.__filtered_done.update(set(new_done).intersection(self.__sweeps))
            inprogress_file.seek(0, os.SEEK_SET)
            try:
                self.__inprogress = pickle.load(inprogress_file)
            except:
                inprogress_file.truncate(0)
                self.__inprogress.clear()
            self.__remaining.difference_update(self.__inprogress)
            self.__filtered_inprogress.intersection_update(self.__inprogress)
        else:
            self.__nolock_full_update(done_file, inprogress_file)

    def update(self):
        """Update incrementaly the ParamSweeper state from disk

        fast, except if done file has been truncated or deleted. In
        this case, will trigger a full_reload.
        """
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)

    def __str__(self):
        with self.__lock:
            return "%s <%i total, %i done, %i skipped, %i in progress, %i remaining>" % (
                self.__name,
                len(self.__sweeps),
                len(self.__filtered_done),
                len(self.__filtered_skipped),
                len(self.__filtered_inprogress),
                len(self.__remaining))

    def get_sweeps(self):
        """Returns the iterable of what to iterate on"""
        return self.__sweeps.copy()

    def get_skipped(self):
        """returns an iterable of current *skipped* elements

        The returned iterable is a copy (safe to use without fearing
        concurrent mutations by another thread).
        """
        return self.__filtered_skipped.copy()

    def get_remaining(self):
        """returns an iterable of current remaining *todo* elements

        The returned iterable is a copy (safe to use without fearing
        concurrent mutations by another thread).
        """
        return self.__remaining.copy()

    def get_inprogress(self):
        """returns an iterable of elements currently processed (which were obtained by a call to `execo_engine.sweep.ParamSweeper.get_next`, not yet marked *done* or *skipped*)

        The returned iterable is a copy (safe to use without fearing
        concurrent mutations by another thread).
        """
        return self.__filtered_inprogress.copy()

    def get_done(self):
        """returns an iterable of currently *done* elements

        The returned iterable is a copy (safe to use without fearing
        concurrent mutations by another thread).
        """
        return self.__filtered_done.copy()

    def get_next(self, filtr = None):
        """Return the next element which is *todo*. Returns None if reached end.

        :param filtr: a filter function. If not None, this filter
          takes the iterable of remaining elements and returns a
          filtered iterable. It can be used to filter out some
          combinations and / or control the order of iteration.
        """
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    remaining = self.__remaining
                    if filtr:
                        remaining = filtr(remaining)
                    try:
                        combination = next(iter(remaining))
                    except StopIteration:
                        logger.trace("%s no new combination", self.__name)
                        logger.trace(self)
                        return None
                    self.__remaining.discard(combination)
                    self.__inprogress.add(combination)
                    self.__filtered_inprogress.add(combination)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s new combination: %s", self.__name, combination)
            logger.trace(self)
            return combination

    def get_next_batch(self, num_combs, filtr = None):
        """Return the next elements which are *todo*.

        :param num_combs: how much combinations to get. An array of
          combinations is returned. The size of the array is <=
          num_combs and is limited by the number of available
          remaining combinations

        :param filtr: a filter function. If not None, this filter
          takes the iterable of remaining elements and returns a
          filtered iterable. It can be used to filter out some
          combinations and / or control the order of iteration.
        """
        combinations = []
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    remaining = self.__remaining
                    if filtr:
                        remaining = filtr(remaining)
                    try:
                        for i in remaining:
                            if num_combs <= 0:
                                break
                            combinations.append(i)
                            num_combs -= 1
                    except StopIteration:
                        logger.trace("%s no new combination", self.__name)
                        logger.trace(self)
                    self.__remaining.difference_update(combinations)
                    self.__inprogress.update(combinations)
                    self.__filtered_inprogress.update(combinations)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s new combinations: %s", self.__name, combinations)
            logger.trace(self)
            return combinations

    def done(self, combination):
        """mark the given element *done*"""
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    self.__remaining.discard(combination)
                    self.__inprogress.discard(combination)
                    self.__filtered_inprogress.discard(combination)
                    self.__done.add(combination)
                    if combination in self.__sweeps:
                        self.__filtered_done.add(combination)
                    done_file.seek(0, os.SEEK_END)
                    pickle.dump(combination, done_file)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s combination done: %s", self.__name, combination)
            logger.trace(self)

    def done_batch(self, combinations):
        """mark the given element(s) *done*"""
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    self.__remaining.difference_update(combinations)
                    self.__inprogress.difference_update(combinations)
                    self.__filtered_inprogress.difference_update(combinations)
                    self.__done.update(combinations)
                    filtered_combinations = set(combinations)
                    filtered_combinations.intersection_update(self.__sweeps)
                    self.__filtered_done.update(filtered_combinations)
                    done_file.seek(0, os.SEEK_END)
                    for combination in combinations:
                        pickle.dump(combination, done_file)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s combinations done: %s", self.__name, combinations)
            logger.trace(self)

    def skip(self, combination):
        """mark the given element *skipped*"""
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    self.__skipped.add(combination)
                    if combination in self.__sweeps:
                        self.__filtered_skipped.add(combination)
                    self.__inprogress.discard(combination)
                    self.__filtered_inprogress.discard(combination)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s combination skipped: %s", self.__name, combination)
            logger.trace(self)

    def skip_batch(self, combinations):
        """mark the given element(s) *skipped*"""
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    self.__skipped.update(combinations)
                    filtered_combinations = set(combinations)
                    filtered_combinations.intersection_update(self.__sweeps)
                    self.__filtered_skipped.update(filtered_combinations)
                    self.__inprogress.difference_update(combinations)
                    self.__filtered_inprogress.difference_update(combinations)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s combinations skipped: %s", self.__name, combinations)
            logger.trace(self)

    def cancel(self, combination):
        """cancel processing of the given combination, but don't mark it as skipped, it comes back in the *todo* queue."""
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    if combination in self.__sweeps:
                        self.__remaining.add(combination)
                    self.__inprogress.discard(combination)
                    self.__filtered_inprogress.discard(combination)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s combination cancelled: %s", self.__name, combination)
            logger.trace(self)

    def cancel_batch(self, combinations):
        """cancel processing of the given combination(s), but don't mark it/them as skipped, they comes back in the *todo* queue."""
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    filtered_combinations = set(combinations)
                    filtered_combinations.intersection_update(self.__sweeps)
                    self.__remaining.update(filtered_combinations)
                    self.__inprogress.difference_update(combinations)
                    self.__filtered_inprogress.difference_update(combinations)
                    inprogress_file.truncate(0)
                    pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s combinations cancelled: %s", self.__name, combinations)
            logger.trace(self)

    def reset(self, reset_inprogress = False):
        """reset container: iteration will start from beginning, state *skipped* are forgotten, state *done* are **not** forgotten.

        :param reset_inprogress: default False. If True, state
          *inprogress* is also reset.
        """
        with self.__lock:
            with _openlock(os.path.join(self.__persistence_dir, "done")) as done_file:
                with _openlock(os.path.join(self.__persistence_dir, "inprogress")) as inprogress_file:
                    self.__nolock_update(done_file, inprogress_file)
                    self.__remaining.update(self.__filtered_skipped)
                    self.__skipped.clear()
                    self.__filtered_skipped.clear()
                    if reset_inprogress:
                        self.__inprogress.clear()
                        self.__filtered_inprogress.clear()
                        inprogress_file.truncate(0)
                        pickle.dump(self.__inprogress, inprogress_file)
            logger.trace("%s reset", self.__name)
            logger.trace(self)

    def stats(self):
        """Atomically return the tuple (sweeps, remaining, skipped, inprogress, done)"""
        with self.__lock:
            sweeps = self.get_sweeps()
            remaining = self.get_remaining()
            skipped = self.get_skipped()
            inprogress = self.get_inprogress()
            done = self.get_done()
            return (sweeps,
                    remaining,
                    skipped,
                    inprogress,
                    done)

def sweep_stats(stats):
    """taking stats tuple returned by `execo_engine.sweep.ParamSweeper.stats`, and if the ParamSweeper sweeps are in the format output by `execo_engine.sweep.sweep`, returns a dict detailing number and ratios of remaining, skipped, done, inprogress combinations per combination parameter value."""

    def count(combs):
        counts = dict()
        for comb in combs:
            for k in comb:
                if k not in counts:
                    counts[k] = dict()
                if comb[k] not in counts[k]:
                    counts[k][comb[k]] = 0
                counts[k][comb[k]] += 1
        return counts

    (sweeps,
     remaining,
     skipped,
     inprogress,
     done) = stats
    ctotal = count(sweeps)
    cremaining = count(remaining)
    cskipped = count(skipped)
    cinprogress = count(inprogress)
    cdone = count(done)
    remaining_ratio = dict()
    skipped_ratio = dict()
    inprogress_ratio = dict()
    done_ratio = dict()
    for k1 in ctotal:
        remaining_ratio[k1] = dict()
        skipped_ratio[k1] = dict()
        inprogress_ratio[k1] = dict()
        done_ratio[k1] = dict()
        for k2 in ctotal[k1]:
            if k1 in cremaining and k2 in cremaining[k1]:
                r = cremaining[k1][k2]
            else:
                r = 0
            remaining_ratio[k1][k2] = float(r) / float(ctotal[k1][k2])
            if k1 in cskipped and k2 in cskipped[k1]:
                s = cskipped[k1][k2]
            else:
                s = 0
            skipped_ratio[k1][k2] = float(s) / float(ctotal[k1][k2])
            if k1 in cinprogress and k2 in cinprogress[k1]:
                i = cinprogress[k1][k2]
            else:
                i = 0
            inprogress_ratio[k1][k2] = float(i) / float(ctotal[k1][k2])
            if k1 in cdone and k2 in cdone[k1]:
                d = cdone[k1][k2]
            else:
                d = 0
            done_ratio[k1][k2] = float(d) / float(ctotal[k1][k2])
    return {
        "total": ctotal,
        "remaining": cremaining,
        "remaining_ratio": remaining_ratio,
        "skipped": cskipped,
        "skipped_ratio": skipped_ratio,
        "inprogress": cinprogress,
        "inprogress_ratio": inprogress_ratio,
        "done": cdone,
        "done_ratio": done_ratio
        }
